import re
from typing import Pattern, Match, List, Tuple

from .exceptions import SGFParserError
from .utils import convert_control_chars
from .collection import SGFCollection
from .property_value import SGFPropertyValue
from .node import SGFNode
from .sequence import SGFSequence
from .game_tree import SGFGameTree

reGameTreeStart = re.compile(r"\s*\(")
reGameTreeEnd = re.compile(r"\s*\)")
reNodeStart = re.compile(r"\s*;")
rePropLabel = re.compile(r"\s*([a-zA-Z]+)")
rePropValueStart = re.compile(r"\s*\[")
rePropValueEnd = re.compile(r"]")
reEscape = re.compile(r"\\[]\\]")
reLineBreak = re.compile(r"(\r\n?|\n\r?)*")


class SGFParser:
    def __init__(self, data: str, index: int = 0):
        self.data = data
        self.index = index

    def _match(self, pattern: Pattern) -> Match:
        return pattern.match(self.data, self.index)

    def _search(self, pattern: Pattern) -> Match:
        return pattern.search(self.data, self.index)

    def parse_collection(self) -> SGFCollection:
        """
        Parse multiple SGFGameTrees into SGFCollection.

        Called when "(" encountered.
        Finishes when last ")" encountered.
        """
        game_trees = self.parse_game_trees()

        if not game_trees:
            raise SGFParserError("Expecting SGFCollection", self.data, self.index)

        if self.index != len(self.data) and not self.data[self.index :].isspace():
            raise SGFParserError("Extra data", self.data, self.index)

        return SGFCollection(game_trees)

    def parse_game_trees(self) -> List[SGFGameTree]:
        """
        Parses multiple SGFGameTrees.

        Called when "(" encountered.
        Finishes when last ")" encountered.
        """
        trees = []
        try:
            while True:
                trees.append(self.parse_game_tree())
        except SGFParserError:
            return trees

    def parse_game_tree(self) -> SGFGameTree:
        """
        Parses single SGFGameTree, which contains multiple SGFNodes and SGFGameTrees (variations).

        SGFGameTree example:
            - (;B[dd])
            - (;B[dd];W[pp](;B[dp];W[pd])(;B[pd];W[dp]))
        Called when "(" encountered.
        Finishes when ")" encountered.
        """
        match = self._match(reGameTreeStart)
        if not match:
            raise SGFParserError("Expecting SGFGameTree", self.data, self.index)

        # consume "("
        self.index = match.end()

        sequence = self.parse_sequence()
        variations = self.parse_game_trees()

        match = self._match(reGameTreeEnd)

        if not match:
            raise SGFParserError("Unterminated SGFGameTree", self.data, self.index)

        # consume ")"
        self.index = match.end()

        return SGFGameTree(sequence, variations)

    def parse_sequence(self) -> SGFSequence:
        nodes = []
        try:
            while True:
                nodes.append(self.parse_node())
        except SGFParserError:
            if not nodes:
                raise SGFParserError("Expecting SGFSequence", self.data, self.index)
        return SGFSequence(nodes)

    def parse_node(self) -> SGFNode:
        """
        Parses single SGFNode, which contains multiple SGFPropertys.

        Called when ";" encountered.
        SGFNode ends when encountered one of the following:
            ";" - next SGFNode starts
            "(" - variation SGFGameTree starts
            ")" - SGFGameTree ends
        """
        match = self._match(reNodeStart)

        if not match:
            raise SGFParserError("Expecting SGFNode", self.data, self.index)

        # consume ";"
        self.index = match.end()

        # parse SGFProperties till ";", "(" or ")"
        props = {}
        try:
            while True:
                prop_label, prop_values = self.parse_property()
                props[prop_label] = prop_values
        except SGFParserError:
            pass

        return SGFNode(props)

    def parse_property(self) -> Tuple[str, SGFPropertyValue]:
        """
        Parses single SGFProperty.

        Called when SGFProperty label encountered.
        Finishes when last SGFProperty value is parsed.
        """
        match = self._match(rePropLabel)

        if not match:
            raise SGFParserError("Expecting SGFProperty", self.data, self.index)

        # consume SGFProperty label
        self.index = match.end()

        prop_label = match.group(0)
        values = []
        try:
            while True:
                values.append(self.parse_prop_value())
        except SGFParserError as err:
            if not values:
                raise err

        return prop_label, SGFPropertyValue(values)

    def parse_prop_value(self) -> str:
        """
        Parses single SGFProperty value.

        Called when "[" encountered.
        Finishes when matching "]" encountered.

        Skips all line breaks and unescapes "\\]".
        """
        match = self._match(rePropValueStart)

        if not match:
            raise SGFParserError("Expecting SGFProperty value", self.data, self.index)

        # consume "["
        self.index = match.end()

        data = ""
        while True:
            # remove line breaks
            match_break = self._match(reLineBreak)
            if match_break:
                self.index = match_break.end()

            match_end = self._search(rePropValueEnd)
            if not match_end:
                raise SGFParserError(
                    "Unterminated SGFProperty value", self.data, self.index
                )

            match_escape = self._search(reEscape)

            if not match_escape or match_escape.end() > match_end.end():
                # no more escaped characters in the SGFProperty value
                break

            # add contents of SGFProperty without `\\`
            data += (
                self.data[self.index : match_escape.start()]
                + self.data[match_escape.end() - 1]
            )
            self.index = match_escape.end()

        # add contents of SGFProperty value
        data += self.data[self.index : match_end.start()]

        # consume "]"
        self.index = match_end.end()

        return convert_control_chars(data)



from typing import TYPE_CHECKING, List

from .exceptions import SGFSequenceError
from .node import SGFNode

if TYPE_CHECKING:
    from .types import SGFNodeType


class SGFSequence(List[SGFNode]):
    def __init__(self, sequence: List["SGFNodeType"]):
        if not sequence:
            raise SGFSequenceError("Expected at least one SGFNode in SGFSequence.")
        super().__init__(map(SGFNode, sequence))

    @property
    def sgf(self) -> str:
        return "".join(node.sgf for node in self)

    def __repr__(self):
        return f"SGFSequence({self.sgf})"

    def copy(self) -> "SGFSequence":
        return SGFSequence([node.copy() for node in self])

    def cut(self, index: int) -> "SGFSequence":
        removed = []
        while len(self) > index:
            removed.insert(0, self.pop())
        return SGFSequence(removed)



from copy import deepcopy
from typing import TYPE_CHECKING, Set, List, Union

from .enums import Player, Location
from .exceptions import SGFBoardError

if TYPE_CHECKING:
    from .types import SGFCoordinateType


class SGFBoard:
    def __init__(
        self,
        shape: "SGFCoordinateType",
        player: Union[str, Player],
        data: List[List[int]] = None,
        allow_super_ko: bool = True,
        allow_suicide: bool = False,
    ):
        rows, cols = shape

        self.shape = shape
        self.player = Player(player)
        self.captured = {Player.BLACK: 0, Player.WHITE: 0}
        self.history = []
        self.allow_super_ko = allow_super_ko
        self.allow_suicide = allow_suicide

        if not (0 < rows <= 52 and 0 < cols <= 52):
            raise SGFBoardError("Board dimensions should be between 1 and 52.")

        if data:
            if len(data) != rows or any(len(row) != cols for row in data):
                raise SGFBoardError(f"Data should be of shape {shape}.")
            self.data = [[Location(point) for point in row] for row in data]
        else:
            self.data = [[Location.EMPTY for _ in range(cols)] for _ in range(rows)]

    def __str__(self, spaces: int = 1):
        width, height = self.shape
        s = ""
        for x in range(height):
            for y in range(width):
                s += self.data[y][x].pretty() + " " * spaces
            s += "\n"
        return s

    def __repr__(self):
        return f"SGFBoard({self})"

    def push(self):
        """Save position to history"""
        state = deepcopy(self.data), self.player, self.captured.copy()
        self.history.append(state)

    def undo(self):
        """Load previous position"""
        self.data, self.player, self.captured = self.history.pop()

    def _is_valid_coord(self, coord: "SGFCoordinateType") -> bool:
        x, y = coord
        rows, cols = self.shape
        return 0 <= x < rows and 0 <= y < cols

    def _get_loc(self, coord: "SGFCoordinateType") -> Location:
        if self._is_valid_coord(coord):
            x, y = coord
            return self.data[x][y]
        raise SGFBoardError("Wrong coordinate")

    def _set_coord(self, coord: "SGFCoordinateType", color: Location):
        if self._is_valid_coord(coord):
            x, y = coord
            self.data[x][y] = color

    def _get_adjacent(self, coord: "SGFCoordinateType") -> List["SGFCoordinateType"]:
        x, y = coord
        for adjacent in [(x - 1, y), (x + 1, y), (x, y - 1), (x, y + 1)]:
            if self._is_valid_coord(adjacent):
                yield adjacent

    def _get_group(self, coord: "SGFCoordinateType") -> Set["SGFCoordinateType"]:
        loc = self._get_loc(coord)

        unexplored = {coord}
        group = set()
        while unexplored:
            coord = unexplored.pop()
            group.add(coord)

            for coord in self._get_adjacent(coord):
                if coord not in group and self._get_loc(coord) is loc:
                    unexplored.add(coord)

        return group

    def _is_alive(self, group: Set["SGFCoordinateType"]) -> bool:
        return any(
            self._get_loc(coord) is Location.EMPTY
            for stone in group
            for coord in self._get_adjacent(stone)
        )

    def _kill_group(self, coord: "SGFCoordinateType", player: Player) -> int:
        if self._get_loc(coord) != player.loc():
            return 0

        group = self._get_group(coord)

        if self._is_alive(group):
            return 0

        for coord in group:
            self._set_coord(coord, Location.EMPTY)

        return len(group)

    def _capture_stones(self, coord: "SGFCoordinateType"):
        for adjacent in self._get_adjacent(coord):
            self.captured[self.player] += self._kill_group(adjacent, -self.player)

    def _check_point(self, coord: "SGFCoordinateType"):
        if self._get_loc(coord) is not Location.EMPTY:
            raise SGFBoardError("Not empty")

    def _check_suicide(self, coord: "SGFCoordinateType"):
        group = self._get_group(coord)
        if not self._is_alive(group):
            raise SGFBoardError("Suicide")

    def _check_ko(self):
        if len(self.history) > 1 and self.history[-2][0] == self.data:
            raise SGFBoardError("Ko")

    def _check_super_ko(self):
        if any(data == self.data for data, *_ in self.history[-3::-1]):
            raise SGFBoardError("Super-Ko")

    def _move(self, coord: "SGFCoordinateType"):
        if not coord:
            return

        self._check_point(coord)
        self._set_coord(coord, self.player.loc())
        self._capture_stones(coord)

        if self.allow_suicide:
            self.captured[-self.player] += self._kill_group(coord, self.player)
        else:
            self._check_suicide(coord)

        if self.allow_super_ko:
            self._check_ko()
        else:
            self._check_super_ko()

        self.player = -self.player

    def move(self, coord: "SGFCoordinateType" = None):
        self.push()
        try:
            self._move(coord)
        except SGFBoardError as err:
            self.undo()
            raise SGFBoardError(f"Illegal move: {err}.") from err



from typing import TYPE_CHECKING, List, Iterable

from .game_tree import SGFGameTree

if TYPE_CHECKING:
    from .types import SGFGameTreeType


class SGFCollection(List[SGFGameTree]):
    def __init__(self, data: Iterable["SGFGameTreeType"]):
        super().__init__([SGFGameTree(*tree) for tree in data])

    @property
    def sgf(self):
        return "\n\n".join(tree.sgf for tree in self)

    def __repr__(self):
        return f"SGFCollection({self.sgf})"

    def pretty(self):
        return "\n\n".join([tree.pretty() for tree in self])

    def copy(self) -> "SGFCollection":
        return SGFCollection([tree.copy() for tree in self])



from datetime import date

from . import (
    SGFBoard,
    SGFCursor,
    SGFGameTree,
    SGFNode,
    SGFPropertyValue,
)
from .types import SGFCoordinateType
from .enums import Player, GoRules
from .exceptions import SGFCursorError
from .utils import coord_to_point, point_to_coord


def make_root_node(
    shape: SGFCoordinateType, player: Player, rules: str, komi: float
) -> SGFNode:
    cols, rows = shape

    return SGFNode(
        {
            "GM": ["1"],
            "FF": ["4"],
            "CA": ["UTF-8"],
            "AP": ["sgflib:0.1.0"],
            "ST": ["2"],
            "RU": [rules],
            "SZ": [f"{rows}"] if rows == cols else [f"{rows}:{cols}"],
            "KM": [str(int(komi) + 0.5 if komi % 1 else 0)],
            "DT": [date.today().isoformat()],
            "PL": [player.value],
        }
    )


def make_game_tree(
    shape: SGFCoordinateType, player: Player, rules: str, komi: float
) -> SGFGameTree:
    return SGFGameTree([make_root_node(shape, player, rules, komi)])


class SGFKifu:
    def __init__(
        self,
        shape: SGFCoordinateType = (19, 19),
        player: Player = Player.BLACK,
        rules: GoRules = GoRules.JAPANESE,
        komi: float = 6.5,
        game_tree: SGFGameTree = None,
    ):
        self.board = SGFBoard(shape, player)

        if not game_tree:
            game_tree = make_game_tree(shape, player, rules, komi)

        self.cursor = SGFCursor(game_tree)

    @property
    def player(self) -> Player:
        return self.board.player

    @property
    def sgf(self):
        return self.cursor.root_tree.sgf

    def _node_move(self) -> SGFPropertyValue:
        return self.cursor.node.get(self.player.value)

    def next(self, variation: int = 0):
        self.cursor.next(variation)
        self.board.move(point_to_coord(sorted(self._node_move())[0]))

    def previous(self):
        self.board.undo()
        self.cursor.previous()

    def _get_next_moves(self):
        moves = {}
        variation = 0
        while True:
            try:
                self.cursor.next(variation)
            except SGFCursorError:
                break

            moves[variation] = self._node_move()
            self.cursor.previous()
            variation += 1

        return moves

    def play(self, coord: SGFCoordinateType):
        point = coord_to_point(coord)
        for index, move in self._get_next_moves().items():
            if move == {point}:
                return self.next(index)

        self.board.move(coord)
        tree = ([{(-self.player).value: [point]}], [])
        tree_index = self.cursor.insert(tree)
        self.cursor.next(tree_index)



from typing import TYPE_CHECKING, Set

from .exceptions import SGFPropertyValueError
from .utils import escape_text

if TYPE_CHECKING:
    from .types import SGFPropertyValueType


class SGFPropertyValue(Set[str]):
    def __init__(self, data: "SGFPropertyValueType"):
        if not data:
            raise SGFPropertyValueError("Cannot be empty")
        super().__init__(data)

    @property
    def sgf(self) -> str:
        return "[" + "][".join(sorted(map(escape_text, self))) + "]"

    def __repr__(self):
        return f"SGFPropertyValue({self.sgf})"

    def copy(self) -> "SGFPropertyValue":
        return SGFPropertyValue(self)

    def clear(self):
        raise SGFPropertyValueError("Cannot clear")

    def pop(self) -> str:
        if len(self) == 1:
            raise SGFPropertyValueError("Cannot pop last element")
        return super().pop()

    def remove(self, element: str):
        if self == {element}:
            raise SGFPropertyValueError("Cannot remove last element")
        return super().remove(element)

    def discard(self, element: str):
        if self == {element}:
            raise SGFPropertyValueError("Cannot discard last element")
        return super().discard(element)



from .property_value import SGFPropertyValue
from .node import SGFNode
from .sequence import SGFSequence
from .game_tree import SGFGameTree
from .collection import SGFCollection
from .parser import SGFParser
from .cursor import SGFCursor
from .board import SGFBoard
from .kifu import SGFKifu



import re
from string import ascii_letters
from typing import TYPE_CHECKING

from .exceptions import SGFCoordinateError, SGFPointError

if TYPE_CHECKING:
    from .types import SGFCoordinateType

reCharsToEscape = re.compile(r"([]\\])")  # characters that need to be \escaped


def convert_control_chars(s: str) -> str:
    """Converts control characters in [text] to spaces. Override for variant behaviour."""
    return s.translate(
        str.maketrans(
            "\000\001\002\003\004\005\006\007\010\011"
            "\013\014\016\017\020\021\022\023\024\025"
            "\026\027\030\031\032\033\034\035\036\037",
            " " * 30,
        )
    )


def escape_text(s: str):
    return reCharsToEscape.sub(r"\\\1", s)


def coord_to_point(coord: "SGFCoordinateType") -> str:
    x, y = coord
    if not (0 <= x < 52 and 0 <= y < 52):
        raise SGFCoordinateError(f"Invalid SGFCoordinate: {coord}.")
    return ascii_letters[x] + ascii_letters[y]


def point_to_coord(point: str) -> "SGFCoordinateType":
    x, y = point
    if x not in ascii_letters or y not in ascii_letters:
        raise SGFPointError(f"Invalid SGFPoint: {point}.")
    return ascii_letters.index(x), ascii_letters.index(y)



from enum import Enum, IntEnum


class Location(IntEnum):
    EMPTY = 0
    BLACK = 1
    WHITE = -1

    def pretty(self):
        if self is self.BLACK:
            return "⬤"
        if self is self.WHITE:
            return "◯"
        return "+"


class Player(str, Enum):
    BLACK = "B"
    WHITE = "W"

    def __neg__(self) -> "Player":
        return Player.BLACK if self is Player.WHITE else Player.WHITE

    def loc(self) -> "Location":
        return Location.BLACK if self is Player.BLACK else Location.WHITE


class GoRules(str, Enum):
    JAPANESE = "Japanese"
    CHINESE = "Chinese"



from typing import TYPE_CHECKING, List, Tuple, Iterable

from .sequence import SGFSequence
from .exceptions import SGFGameTreeError

if TYPE_CHECKING:
    from .types import SGFSequenceType, SGFGameTreeType


class SGFGameTree(Tuple[SGFSequence, List["SGFGameTree"]]):
    def __new__(
        cls,
        sequence: "SGFSequenceType",
        variations: Iterable["SGFGameTreeType"] = (),
    ):
        sequence = SGFSequence(sequence)
        variations = [SGFGameTree(*tree) for tree in variations]
        return super().__new__(cls, (sequence, variations))

    def __init__(self, *_, **__):
        self.sequence, self.variations = self

    def __repr__(self):
        return f"SGFGameTree({self.sgf})"

    @property
    def sgf(self):
        return (
            "("
            + self.sequence.sgf
            + "".join(tree.sgf for tree in self.variations)
            + ")"
        )

    def __eq__(self, other: "SGFGameTreeType"):
        return super().__eq__(SGFGameTree(*other))

    def copy(self) -> "SGFGameTree":
        return SGFGameTree(
            sequence=self.sequence.copy(),
            variations=[tree.copy() for tree in self.variations],
        )

    def pretty(self, offset: int = 0, indent: int = 2):
        s = " " * offset + "(\n" + " " * (offset + indent) + self.sequence.sgf
        for tree in self.variations:
            s += "\n" + " " * offset + tree.pretty(offset + indent, indent)
        return s + "\n" + " " * offset + ")"

    def insert(self, tree: "SGFGameTreeType", index: int) -> int:
        tree = SGFGameTree(*tree)
        if index < 1 or index > len(self.sequence):
            raise SGFGameTreeError(f"Cannot insert SGFGameTree at index={index}.")
        if index < len(self.sequence):
            main_tree = SGFGameTree(
                sequence=self.sequence.cut(index),
                variations=[tree.copy() for tree in self.variations],
            )
            self.variations.clear()
            self.variations.extend([main_tree, tree])
            return 1
        if self.variations:
            self.variations.append(tree)
            return len(self.variations) - 1

        self.sequence.extend(tree.sequence)
        self.variations.extend(tree.variations)
        return 0

    def cut_tree(self, index: int) -> "SGFGameTree":
        if index < 1 or index > len(self.sequence) - 1:
            raise SGFGameTreeError(f"Cannot cut SGFSequence at index={index}.")
        sequence = self.sequence.cut(index)
        variations = self.variations.copy()
        self.variations.clear()
        return SGFGameTree(sequence, variations)

    def cut_variation(self, index: int):
        if not (0 <= index < len(self.variations)):
            raise SGFGameTreeError(
                f"Cannot cut variation SGFGameTree at index={index}."
            )
        self.variations.pop(index)



from typing import Tuple, List, Iterable, Mapping


SGFPropertyValueType = Iterable[str]
SGFNodeType = Mapping[str, SGFPropertyValueType]
SGFSequenceType = List[SGFNodeType]
SGFGameTreeType = Tuple[SGFSequenceType, List["SGFGameTreeType"]]

SGFCoordinateType = Tuple[int, int]



from typing import TYPE_CHECKING, Mapping, Dict

from .property_value import SGFPropertyValue

if TYPE_CHECKING:
    from .types import SGFNodeType, SGFPropertyValueType


class SGFNode(Dict[str, SGFPropertyValue]):
    def __init__(self, data: "SGFNodeType" = None, **kwargs: "SGFPropertyValueType"):
        super().__init__()
        self.update(data, **kwargs)

    @property
    def sgf(self) -> str:
        sgf_string = ";"
        for prop_label, prop_value in sorted(self.items()):
            sgf_string += prop_label + prop_value.sgf
        return sgf_string

    def __repr__(self):
        return f"SGFNode({self.sgf})"

    def copy(self) -> "SGFNode":
        return SGFNode({key: value.copy() for key, value in self.items()})

    def update(
        self,
        data: "SGFNodeType" = None,
        **kwargs: "SGFPropertyValueType",
    ):
        data = data or []
        if isinstance(data, Mapping):
            data = data.items()
        data = {label.upper(): SGFPropertyValue(values) for label, values in data}

        kwargs = {
            label.upper(): SGFPropertyValue(values) for label, values in kwargs.items()
        }
        super().update(data, **kwargs)

    def setdefault(
        self, key: str, default: "SGFPropertyValueType" = ...
    ) -> SGFPropertyValue:
        return super().setdefault(key, SGFPropertyValue(default))

    def __setitem__(self, key: str, values: "SGFPropertyValueType"):
        return super().__setitem__(key, SGFPropertyValue(values))



class SGFParserError(ValueError):
    def __init__(self, msg: str, data: str, index: int):
        lineno = data.count("\n", 0, index) + 1
        colno = index - data.rfind("\n", 0, index)
        errmsg = f"{msg}: line {lineno} column {colno} (char {index})"

        super().__init__(errmsg)
        self.msg = msg
        self.data = data
        self.index = index
        self.lineno = lineno
        self.colno = colno


class SGFPropertyValueError(Exception):
    pass


class SGFGameTreeError(Exception):
    pass


class SGFSequenceError(Exception):
    pass


class SGFCursorError(Exception):
    pass


class SGFBoardError(Exception):
    pass


class SGFCoordinateError(Exception):
    pass


class SGFPointError(Exception):
    pass



from typing import TYPE_CHECKING, List

from . import SGFNode, SGFGameTree
from .exceptions import SGFCursorError

if TYPE_CHECKING:
    from .types import SGFGameTreeType


class SGFCursor:
    def __init__(self, tree: "SGFGameTreeType"):
        self.root_tree = self.tree = SGFGameTree(*tree)
        self.index = 0
        self._stack: List[SGFGameTree] = []

    @property
    def node(self) -> SGFNode:
        return self.tree.sequence[self.index]

    def next(self, variation: int = 0) -> SGFNode:
        if self.index + 1 < len(self.tree.sequence):
            if variation != 0:
                raise SGFCursorError("Invalid variation number.")
            self.index += 1
        elif self.tree.variations:
            if variation >= len(self.tree.variations):
                raise SGFCursorError("Invalid variation number.")
            self._stack.append(self.tree)
            self.tree = self.tree.variations[variation]
            self.index = 0
        else:
            raise SGFCursorError("Reached end of SGFGameTree.")

        return self.node

    def previous(self) -> SGFNode:
        if self.index:
            self.index -= 1
        elif self._stack:
            self.tree = self._stack.pop()
            self.index = len(self.tree.sequence) - 1
        else:
            raise SGFCursorError("Reached start of SGFGameTree.")

        return self.node

    def insert(self, tree: "SGFGameTreeType") -> int:
        return self.tree.insert(tree, self.index + 1)
